#!/usr/bin/env python
# -*- coding: utf-8 -*-

from asyncio import gather
from asyncio import sleep
from itertools import groupby

from quart import Blueprint
from quart import flash
from quart import redirect
from quart import render_template
from quart import url_for
from uscpi import TCP

from .db import get_db

home = Blueprint("home", __name__)


async def sendall(host: str, port: int, sequences: list) -> None:
    async with TCP(host=host, port=port, timeout=5) as client:
        for sequence in sequences:
            command = bytes(sequence["scpi"], encoding="utf8")
            await client.write(command + b"\n")
            await sleep(0.1)  # 100 milliseconds


@home.get("/")
async def index():
    rows = get_db().execute("SELECT * FROM state").fetchall()
    return await render_template("home.html", states=rows)


@home.get("/configure/<int:state_id>")
async def configure(state_id: int):
    rows = (
        get_db()
        .execute(
            """
        SELECT
            instrument.host AS host,
            instrument.port AS port,
            command.scpi AS scpi
        FROM sequence
            INNER JOIN command ON command.id = sequence.command_id
            INNER JOIN instrument ON instrument.id = sequence.instrument_id
            INNER JOIN state ON state.id = sequence.state_id
        WHERE state.id = ?
        """,
            (state_id,),
        )
        .fetchall()
    )
    keyfunc = lambda x: (x["host"], x["port"])
    records = sorted(map(dict, rows), key=keyfunc)
    groups = groupby(records, key=keyfunc)
    coros = [sendall(*p, list(s)) for p, s in groups]
    try:
        await gather(*coros)
        await flash(f"Setup complete.", "success")
    except TimeoutError:
        await flash(f"Timeout Error. Verify instrumentation is powered on.", "error")
    except OSError as e:
        await flash(f"{e}", "failed")
    return redirect(url_for("home.index"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from sqlite3 import connect
from tempfile import NamedTemporaryFile
from unittest import IsolatedAsyncioTestCase
from unittest import main

from minstrel import create_app


class SequenceTestCase(IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        cls._resources = Path(__file__).parent
        path = cls._resources / "preload.sql"
        with open(path, mode="r", encoding="utf-8") as f:
            cls._preload = f.read()

    def setUp(self):
        self.db = NamedTemporaryFile()
        self.app = create_app({"TESTING": True, "DATABASE": self.db.name})
        self.client = self.app.test_client()
        self.app.test_cli_runner().invoke(args=["init-db"])
        db = connect(self.db.name)
        db.executescript(self._preload)
        resp = self.app.test_cli_runner().invoke(args=["token"])
        self.token = resp.output.rstrip()

    def tearDown(self):
        self.db.close()

    async def test_create_sequence(self):
        response = await self.client.post(
            f"/api/sequence?token={self.token}",
            form={
                "state_id": 1,
                "instrument_id": 1,
                "command_id": 2,
            },
        )
        self.assertEqual(response.status_code, 201)

    async def test_read_sequence(self):
        response = await self.client.get("/api/sequence/1")
        self.assertEqual(response.status_code, 200)

    async def test_update_sequence(self):
        response = await self.client.put(
            f"/api/sequence/1?token={self.token}",
            form={
                "state_id": 1,
                "instrument_id": 1,
                "command_id": 3,
            },
        )
        self.assertEqual(response.status_code, 201)

    async def test_delete_sequence(self):
        response = await self.client.delete(
            f"/api/sequence/1?token={self.token}",
        )
        self.assertEqual(response.status_code, 200)

    async def test_list_sequences(self):
        response = await self.client.get("/api/sequence")
        self.assertEqual(response.status_code, 200)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from quart import Blueprint
from quart import render_template
from quart import request

from ..db import get_db
from ..slug import slugify
from ..token import token_required

sequence = Blueprint(
    "sequence",
    __name__,
    url_prefix="/api",
)


@sequence.post("/sequence")
@token_required
async def create_sequence():
    """Create sequence."""

    form = (await request.form).copy().to_dict()
    try:
        db = get_db()
        db.execute("PRAGMA foreign_keys = ON")
        db.execute(
            """
            INSERT INTO sequence (
                instrument_id,
                state_id,
                command_id
            ) VALUES (
                :instrument_id,
                :state_id,
                :command_id
            )
            """,
            form,
        )
        db.commit()
    except db.ProgrammingError:
        return "Missing parameter(s).", 400
    except db.IntegrityError:
        return "Invalid parameter(s).", 400
    return "Sequence successfully created.", 201


@sequence.get("/sequence/<int:id>")
async def read_sequence(id: int):
    """Read sequence."""

    row = (
        get_db()
        .execute(
            "SELECT * FROM sequence WHERE id = ?",
            (id,),
        )
        .fetchone()
    )
    if not row:
        return "Sequence does not exist.", 404
    return dict(row), 200


@sequence.put("/sequence/<int:id>")
@token_required
async def update_sequence(id: int):
    """Update sequence."""

    form = (await request.form).copy().to_dict()
    form["id"] = id
    try:
        db = get_db()
        db.execute("PRAGMA foreign_keys = ON")
        db.execute(
            """
            UPDATE sequence SET
                updated_at = CURRENT_TIMESTAMP,
                instrument_id = :instrument_id,
                state_id = :state_id,
                command_id = :command_id
            WHERE id = :id
            """,
            form,
        )
        db.commit()
    except db.ProgrammingError:
        return "Missing parameter(s).", 400
    except db.IntegrityError:
        return "Invalid parameter(s).", 400
    return "Sequence successfully updated.", 201


@sequence.delete("/sequence/<int:id>")
@token_required
async def delete_sequence(id: int):
    """Delete sequence."""

    db = get_db()
    db.execute("DELETE FROM sequence WHERE id = ?", (id,))
    db.commit()
    return "Sequence successfully deleted.", 200


@sequence.get("/sequence")
async def list_sequences():
    """List sequences."""

    rows = get_db().execute("SELECT * FROM sequence").fetchall()
    if not rows:
        return "Sequences do not exist.", 404
    return list(map(dict, rows)), 200



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from unittest import IsolatedAsyncioTestCase
from unittest import main

from minstrel import create_app


class InitTestCase(IsolatedAsyncioTestCase):
    def setUp(self):
        self.db = "file::memory:?cache=shared"
        self.app = create_app({"TESTING": True, "DATABASE": self.db})
        self.runner = self.app.test_cli_runner()
        self.ctx = self.app.app_context()

    def test_db_close(self):
        result = self.ctx.g.get("db")
        self.assertIsNone(result, result)

    def test_db_init_command(self):
        response = self.runner.invoke(args=["init-db"])
        self.assertIsInstance(response.output, str)

    def test_token_command(self):
        response = self.runner.invoke(args=["token"])
        self.assertIsInstance(response.output, str)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime
from pathlib import Path
from sqlite3 import connect
from sqlite3 import PARSE_DECLTYPES
from sqlite3 import register_converter
from sqlite3 import Row

from click import command
from click import echo
from quart import current_app
from quart import g
from quart.cli import with_appcontext


def convert_datetime(value: bytes):
    """
    Convert ISO 8601 datetime to datetime.datetime object.
    """

    return datetime.fromisoformat(value.decode())


def get_db():
    """
    Connect to the application's configured database. The connection
    is unique for each request and will be reused if this is called
    again.
    """

    if not hasattr(g, "db"):
        register_converter("datetime", convert_datetime)
        g.db = connect(
            current_app.config["DATABASE"],
            detect_types=PARSE_DECLTYPES,
        )
        g.db.row_factory = Row
    return g.db


async def close_db(exception=None):
    """
    If this request connected to the database, close the connection.
    """

    db = g.pop("db", None)
    if db is not None:
        db.close()


@command("init-db")
@with_appcontext
def init_db_command() -> None:
    """
    Clear existing data and create new tables.
    """

    db = get_db()
    root_path = Path(current_app.root_path)
    with open(root_path / "schema.sql") as file:
        db.executescript(file.read())
    echo("Database initialized.")


def init_app(app) -> None:
    """
    Register database functions with the Quart app. This is called by
    the application factory.
    """

    app.teardown_appcontext(close_db)
    app.cli.add_command(init_db_command)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime
from datetime import timedelta
from datetime import timezone
from functools import wraps

from click import argument
from click import command
from click import echo
from jwt import decode
from jwt import encode
from quart import current_app
from quart import request
from quart.cli import with_appcontext


def token_required(view):
    """
    Token required decorator function.
    """

    @wraps(view)
    async def wrapper(*args, **kwargs):
        if not request.args.get("token", None):
            return "Token required.", 401
        try:
            decode(
                request.args["token"],
                current_app.config["SECRET_KEY"],
                algorithms=["HS256"],
            ).get("confirm")
        except Exception as error:
            return f"{error}", 401
        return await view(*args, **kwargs)

    return wrapper


@command("token")
@argument("expires_in", type=int, default=300)
@with_appcontext
def init_token_command(expires_in: int) -> None:
    """
    Initialize and generate new token key given a specified
    expiration timeframe where the default expiration time
    is 300 seconds (5 minutes).
    """

    delta = timedelta(seconds=expires_in)
    exp = datetime.now(tz=timezone.utc) + delta
    token = encode(
        payload={"confirm": "42", "exp": exp},
        key=current_app.config["SECRET_KEY"],
        algorithm="HS256",
    )
    echo(f"{token}")


def init_token(app):
    """
    Add command to Quart application instance to initialize
    and generate a new token key.
    """

    app.cli.add_command(init_token_command)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from quart import Blueprint
from quart import flash
from quart import redirect
from quart import render_template
from quart import request
from quart import url_for

from .db import get_db
from .api.command import create_command as create_command_api
from .api.command import update_command as update_command_api
from .api.command import delete_command as delete_command_api
from .api.instrument import create_instrument as create_instrument_api
from .api.instrument import update_instrument as update_instrument_api
from .api.instrument import delete_instrument as delete_instrument_api
from .api.sequence import create_sequence as create_sequence_api
from .api.sequence import update_sequence as update_sequence_api
from .api.sequence import delete_sequence as delete_sequence_api
from .api.state import create_state as create_state_api
from .api.state import update_state as update_state_api
from .api.state import delete_state as delete_state_api

settings = Blueprint("settings", __name__)


@settings.get("/settings")
async def index():
    states = get_db().execute("SELECT * FROM state").fetchall()
    instruments = get_db().execute("SELECT * FROM instrument").fetchall()
    commands = get_db().execute("SELECT * FROM command").fetchall()
    sequences = (
        get_db()
        .execute(
            """
        SELECT
            sequence.id AS id,
            sequence.command_id AS command_id,
            sequence.instrument_id AS instrument_id,
            sequence.state_id AS state_id,
            command.title AS command_title,
            instrument.title AS instrument_title,
            state.title AS state_title
        FROM sequence
            INNER JOIN state ON state.id = sequence.state_id
            INNER JOIN instrument ON instrument.id = sequence.instrument_id
            INNER JOIN command ON command.id = sequence.command_id
        """
        )
        .fetchall()
    )
    return await render_template(
        "settings.html",
        commands=commands,
        states=states,
        instruments=instruments,
        sequences=sequences,
    )


@settings.route("/settings/instrument/create", methods=("GET", "POST"))
async def create_instrument():
    if request.method == "POST":
        response = await create_instrument_api.__wrapped__()
        if response[1] < 300:
            return redirect(url_for(".index"))
        await flash(response[0], "error")
    return await render_template("create_instrument.html")


@settings.route("/settings/instrument/<int:id>/update", methods=("GET", "POST"))
async def update_instrument(id: int):
    if request.method == "POST":
        response = await update_instrument_api.__wrapped__(id)
        if response[1] < 300:
            return redirect(url_for(".index"))
        await flash(response[0], "error")
    row = get_db().execute("SELECT * FROM instrument WHERE id = ?", (id,)).fetchone()
    return await render_template("update_instrument.html", row=row)


@settings.route("/settings/instrument/<int:id>/delete", methods=("GET",))
async def delete_instrument(id: int):
    response = await delete_instrument_api.__wrapped__(id)
    if response[1] >= 300:
        await flash(response[0], "error")
    return redirect(url_for(".index"))


@settings.route("/settings/state/create", methods=("GET", "POST"))
async def create_state():
    if request.method == "POST":
        response = await create_state_api.__wrapped__()
        if response[1] < 300:
            return redirect(url_for(".index"))
        await flash(response[0], "error")
    return await render_template("create_state.html")


@settings.route("/settings/state/<int:id>/update", methods=("GET", "POST"))
async def update_state(id: int):
    if request.method == "POST":
        response = await update_state_api.__wrapped__(id)
        if response[1] < 300:
            return redirect(url_for(".index"))
        await flash(response[0], "error")
    row = get_db().execute("SELECT * FROM state WHERE id = ?", (id,)).fetchone()
    return await render_template("update_state.html", row=row)


@settings.route("/settings/state/<int:id>/delete", methods=("GET",))
async def delete_state(id: int):
    response = await delete_state_api.__wrapped__(id)
    if response[1] >= 300:
        await flash(response[0], "error")
    return redirect(url_for(".index"))


@settings.route("/settings/command/create", methods=("GET", "POST"))
async def create_command():
    if request.method == "POST":
        response = await create_command_api.__wrapped__()
        if response[1] < 300:
            return redirect(url_for(".index"))
        await flash(response[0], "error")
    return await render_template("create_command.html")


@settings.route("/settings/command/<int:id>/update", methods=("GET", "POST"))
async def update_command(id: int):
    if request.method == "POST":
        response = await update_command_api.__wrapped__(id)
        if response[1] < 300:
            return redirect(url_for(".index"))
        await flash(response[0], "error")
    row = get_db().execute("SELECT * FROM command WHERE id = ?", (id,)).fetchone()
    return await render_template("update_command.html", row=row)


@settings.route("/settings/command/<int:id>/delete", methods=("GET",))
async def delete_command(id: int):
    response = await delete_command_api.__wrapped__(id)
    if response[1] >= 300:
        await flash(response[0], "error")
    return redirect(url_for(".index"))


@settings.route("/settings/sequence/create", methods=("GET", "POST"))
async def create_sequence():
    if request.method == "POST":
        response = await create_sequence_api.__wrapped__()
        if response[1] < 300:
            return redirect(url_for(".index"))
        await flash(response[0], "error")
    commands = get_db().execute("SELECT * FROM command").fetchall()
    states = get_db().execute("SELECT * FROM state").fetchall()
    instruments = get_db().execute("SELECT * FROM instrument").fetchall()
    return await render_template(
        "create_sequence.html",
        instruments=instruments,
        states=states,
        commands=commands,
    )


@settings.route("/settings/sequence/<int:id>/update", methods=("GET", "POST"))
async def update_sequence(id: int):
    if request.method == "POST":
        response = await update_sequence_api.__wrapped__(id)
        if response[1] < 300:
            return redirect(url_for(".index"))
        await flash(response[0], "error")
    row = get_db().execute("SELECT * FROM sequence WHERE id = ?", (id,)).fetchone()
    commands = get_db().execute("SELECT * FROM command").fetchall()
    states = get_db().execute("SELECT * FROM state").fetchall()
    instruments = get_db().execute("SELECT * FROM instrument").fetchall()
    return await render_template(
        "update_sequence.html",
        row=row,
        instruments=instruments,
        states=states,
        commands=commands,
    )


@settings.route("/settings/sequence/<int:id>/delete", methods=("GET",))
async def delete_sequence(id: int):
    response = await delete_sequence_api.__wrapped__(id)
    if response[1] >= 300:
        await flash(response[0], "error")
    return redirect(url_for(".index"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from os import path
from os import makedirs

from quart import Quart
from quart import render_template

from .api.command import command
from .api.instrument import instrument
from .api.sequence import sequence
from .api.state import state
from .home import home
from .settings import settings
from .db import init_app
from .token import init_token

__version__ = "0.0.1"


async def page_not_found(e):
    return await render_template("404_not_found.html"), 404


def create_app(test_config=None):
    """
    Create and configure an instance of the Quart
    application.
    """

    app = Quart(
        __name__,
        instance_relative_config=True,
    )
    app.config.from_mapping(
        SECRET_KEY="dev",
        MAX_CONTENT_LENGTH=16000000,
        DATABASE=path.join(
            app.instance_path,
            "minstrel.sqlite",
        ),
    )
    if test_config is None:
        app.config.from_pyfile(
            "config.py",
            silent=True,
        )
    else:
        app.config.update(test_config)
    try:
        makedirs(app.instance_path)
    except OSError:
        pass
    init_app(app)
    init_token(app)
    app.register_blueprint(command)
    app.register_blueprint(instrument)
    app.register_blueprint(state)
    app.register_blueprint(sequence)
    app.register_blueprint(home)
    app.register_blueprint(settings)
    app.register_error_handler(404, page_not_found)
    return app



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from quart import Blueprint
from quart import render_template
from quart import request

from ..db import get_db
from ..slug import slugify
from ..token import token_required

instrument = Blueprint(
    "instrument",
    __name__,
    url_prefix="/api",
)


@instrument.post("/instrument")
@token_required
async def create_instrument():
    """Create instrument."""

    form = (await request.form).copy().to_dict()
    try:
        form["slug"] = slugify(form.get("title"))
        db = get_db()
        db.execute("PRAGMA foreign_keys = ON")
        db.execute(
            """
            INSERT INTO instrument (
                slug,
                title,
                host,
                port
            ) VALUES (
                :slug,
                :title,
                :host,
                :port
            )
            """,
            form,
        )
        db.commit()
    except db.ProgrammingError:
        return "Missing parameter(s).", 400
    except db.IntegrityError:
        return "Invalid parameter(s).", 400
    return "Instrument successfully created.", 201


@instrument.get("/instrument/<int:id>")
async def read_instrument(id: int):
    """Read instrument."""

    row = (
        get_db()
        .execute(
            "SELECT * FROM instrument WHERE id = ?",
            (id,),
        )
        .fetchone()
    )
    if not row:
        return "Instrument does not exist.", 404
    return dict(row), 200


@instrument.put("/instrument/<int:id>")
@token_required
async def update_instrument(id: int):
    """Update instrument."""

    form = (await request.form).copy().to_dict()
    form["id"] = id
    try:
        form["slug"] = slugify(form.get("title"))
        db = get_db()
        db.execute("PRAGMA foreign_keys = ON")
        db.execute(
            """
            UPDATE instrument SET
                updated_at = CURRENT_TIMESTAMP,
                slug = :slug,
                title = :title,
                host = :host,
                port = :port
            WHERE id = :id
            """,
            form,
        )
        db.commit()
    except db.ProgrammingError:
        return "Missing parameter(s).", 400
    except db.IntegrityError:
        return "Invalid parameter(s).", 400
    return "Instrument successfully updated.", 201


@instrument.delete("/instrument/<int:id>")
@token_required
async def delete_instrument(id: int):
    """Delete instrument."""

    db = get_db()
    db.execute("DELETE FROM instrument WHERE id = ?", (id,))
    db.commit()
    return "Instrument successfully deleted.", 200


@instrument.get("/instrument")
async def list_instruments():
    """List instruments."""

    rows = get_db().execute("SELECT * FROM instrument").fetchall()
    if not rows:
        return "Instruments do not exist.", 404
    return list(map(dict, rows)), 200



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from re import sub


def slugify(x: str) -> str:
    """
    Convert string into slug.
    """

    x = x.lower().strip()
    x = sub(r"[^\w\s-]", "", x)
    x = sub(r"[\s_-]+", "_", x)
    x = sub(r"^-+|-+$", "", x)
    return x
